class WormholeGuiError(Exception):
    pass


class RespondError(WormholeGuiError):
    """Error that needs to be signalled across the wormhole"""

    def __init__(self, cause):
        self.cause = cause


class RemoteError(WormholeGuiError):
    """Error that was signaled from the other end of the wormhole"""

    pass


class SendTextError(WormholeGuiError):
    """Other side sent message_ack not ok"""

    pass


class SendFileError(WormholeGuiError):
    """Other side sent file_ack not ok"""

    pass


class ReceiveFileError(WormholeGuiError):
    """Transit connection closed before full file was received"""

    pass


class OfferError(WormholeGuiError):
    """Invalid offer received"""

    pass


class DiskSpaceError(WormholeGuiError):
    """Couldn't receive a file due to low disk space"""

    pass


class RefusedError(WormholeGuiError):
    """The file transfer was refused"""

    pass



import os
from pathlib import Path

SHELL_FOLDERS = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders"
DOWNLOADS_GUID = "{374DE290-123F-4565-9164-39C4925E467B}"


def get_download_path_or_cwd():
    download_path = get_download_path()
    if download_path is None or not Path(download_path).exists():
        return Path.cwd()
    else:
        return download_path.resolve()


def get_download_path():
    if os.name == "nt":
        try:
            import winreg

            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, SHELL_FOLDERS) as key:
                return Path(winreg.QueryValueEx(key, DOWNLOADS_GUID)[0])
        except Exception:
            return None
    else:
        return Path.home() / "Downloads"



import json
import logging
import traceback

from PySide2.QtCore import QObject, Slot
import wormhole
from wormhole.cli import public_relay
from wormhole.errors import LonelyError

from .errors import (
    RefusedError,
    RemoteError,
    RespondError,
    SendFileError,
    SendTextError,
)
from .timeout import Timeout
from .transit_protocol_pair import TransitProtocolPair

TIMEOUT_SECONDS = 2
APPID = "lothar.com/wormhole/text-or-file-xfer"


class FileTransferProtocol(QObject):
    def __init__(self, reactor, signals):
        self._reactor = reactor
        self._wormhole = None
        self._is_wormhole_connected = False
        self._transit = None
        self._peer_versions = {}
        self._wormhole_delegate = WormholeDelegate(signals, self._handle_message)
        self._transit_delegate = TransitDelegate(signals)
        self._timeout = Timeout(reactor, TIMEOUT_SECONDS)

        self._signals = signals
        self._signals.versions_received.connect(self._on_versions_received)
        self._signals.wormhole_open.connect(self._on_wormhole_open)
        self._signals.wormhole_closed.connect(self._on_wormhole_closed)
        self._signals.file_transfer_complete.connect(self._on_file_transfer_complete)
        self._signals.respond_error.connect(self._on_respond_error)

    def open(self, code):
        logging.debug("open wormhole")
        assert self._wormhole is None

        self._wormhole = wormhole.create(
            appid=APPID,
            relay_url=public_relay.RENDEZVOUS_RELAY,
            reactor=self._reactor,
            delegate=self._wormhole_delegate,
            versions={"v0": {"mode": "connect"}},
        )

        self._transit = TransitProtocolPair(
            self._reactor, self._wormhole, self._transit_delegate
        )

        if code is None or code == "":
            self._wormhole.allocate_code()
        else:
            self._wormhole.set_code(code)

    def close(self):
        logging.debug("close wormhole")
        if self._wormhole is None:
            self._signals.wormhole_closed.emit()
        else:
            self._transit.close()
            self._wormhole.close()

    def shutdown(self):
        logging.debug("shutdown wormhole")
        if self._wormhole is None:
            self._signals.wormhole_shutdown.emit()
        else:
            if self._is_wormhole_connected and self._peer_supports_connect_mode():
                self.send_command("shutdown")

            self._wormhole_delegate.shutdown()
            self.close()

    @Slot()
    def _on_wormhole_open(self):
        self._is_wormhole_connected = True

    @Slot()
    def _on_wormhole_closed(self):
        self._wormhole = None
        self._is_wormhole_connected = False

    @Slot(dict)
    def _on_versions_received(self, versions):
        self._peer_versions = versions

    @Slot(int, str)
    def _on_file_transfer_complete(self, id, filename):
        if not self._peer_supports_connect_mode():
            self.close()

    @Slot(Exception, str)
    def _on_respond_error(self, exception, traceback):
        self._send_data({"error": str(exception)})
        if isinstance(exception, RefusedError):
            self.close()
            return
        self._signals.error.emit(exception, traceback)

    def _peer_supports_connect_mode(self):
        if "v0" not in self._peer_versions:
            return False
        return self._peer_versions["v0"].get("mode") == "connect"

    def send_message(self, message):
        self._send_data({"offer": {"message": message}})

    def send_command(self, command):
        self._send_data({"command": command})

    def send_file(self, id, file_path):
        self._transit.sender.send_file(id, file_path)

    def receive_file(self, id, dest_path):
        self._transit.receiver.receive_file(id, dest_path)

    def is_sending_file(self):
        return self._transit.sender.is_sending_file

    def is_receiving_file(self):
        return self._transit.sender.is_receiving_file

    def _send_data(self, data):
        assert isinstance(data, dict)
        logging.debug(f"Sending: {data}")
        self._wormhole.send_message(json.dumps(data).encode("utf-8"))

    def _handle_message(self, data_bytes):
        data = json.loads(data_bytes.decode("utf-8"))
        if "error" in data:
            raise RemoteError(data["error"])

        for key, contents in data.items():
            if key == "offer":
                self._handle_offer(contents)

            elif key == "transit":
                self._transit.handle_transit(contents)

            elif key == "command" and contents == "shutdown":
                self._signals.wormhole_shutdown_received.emit()
                self.close()

            elif key == "answer" and "message_ack" in contents:
                result = contents["message_ack"]
                is_ok = result == "ok"
                self._signals.message_sent.emit(is_ok)
                if not is_ok:
                    raise SendTextError(result)
                if not self._peer_supports_connect_mode():
                    self.close()

            elif key == "answer" and "file_ack" in contents:
                result = contents["file_ack"]
                if result == "ok":
                    self._transit.sender.handle_file_ack()
                else:
                    raise SendFileError(result)

            else:
                logging.warning(f"Unexpected data received: {key}: {contents}")

    def _handle_offer(self, offer):
        if "message" in offer:
            self._send_data({"answer": {"message_ack": "ok"}})
            self._signals.message_received.emit(offer["message"])
            if not self._peer_supports_connect_mode():
                self.close()
        else:
            dest = self._transit.receiver.handle_offer(offer)
            self._signals.file_receive_pending.emit(dest.name, dest.final_bytes)


class WormholeDelegate:
    def __init__(self, signals, message_handler):
        self._signals = signals
        self._message_handler = message_handler
        self._shutting_down = False

    def shutdown(self):
        self._shutting_down = True

    def wormhole_got_welcome(self, welcome):
        logging.debug(f"wormhole_got_welcome: {welcome}")

    def wormhole_got_code(self, code):
        logging.debug(f"wormhole_got_code: {code}")
        self._signals.code_received.emit(code)

    def wormhole_got_unverified_key(self, key):
        logging.debug(f"wormhole_got_unverified_key: {key}")

    def wormhole_got_verifier(self, verifier):
        logging.debug(f"wormhole_got_verifier: {verifier}")

    def wormhole_got_versions(self, versions):
        logging.debug(f"wormhole_got_versions: {versions}")
        self._signals.versions_received.emit(versions)
        self._signals.wormhole_open.emit()

    def wormhole_got_message(self, data):
        logging.debug(f"wormhole_got_message: {data}")
        try:
            self._message_handler(data)
        except RespondError as exception:
            self._signals.respond_error.emit(exception.cause, traceback.format.exc())
        except Exception as exception:
            self._signals.error.emit(exception, traceback.format_exc())

    def wormhole_closed(self, result):
        logging.debug(f"wormhole_closed: {repr(result)}")
        if self._shutting_down:
            logging.debug("Emit wormhole_shutdown")
            self._signals.wormhole_shutdown.emit()
        else:
            if isinstance(result, LonelyError):
                pass
            elif isinstance(result, Exception):
                self._signals.error.emit(result, None)
            self._signals.wormhole_closed.emit()


class TransitDelegate:
    def __init__(self, signals):
        self._signals = signals

    def transit_progress(self, id, transferred_bytes, total_bytes):
        self._signals.file_transfer_progress.emit(id, transferred_bytes, total_bytes)

    def transit_complete(self, id, filename):
        logging.debug(f"transit_complete: {id}, {filename}")
        self._signals.file_transfer_complete.emit(id, filename)

    def transit_error(self, exception, traceback=None):
        self._signals.error.emit(exception, traceback)



from .transit_protocol import TransitProtocolReceiver, TransitProtocolSender


class TransitProtocolPair:
    def __init__(self, reactor, wormhole, delegate):
        self.receiver = TransitProtocolReceiver(reactor, wormhole, delegate)
        self.sender = TransitProtocolSender(reactor, wormhole, delegate)

    def handle_transit(self, transit_message):
        if self.sender.awaiting_transit_response:
            # We're waiting for a response, so this is for the sender
            self.sender.handle_transit(transit_message)
        else:
            # We've received a transit message first, so this is for the receiver
            self.receiver.handle_transit(transit_message)

    def close(self):
        self.sender.close()
        self.receiver.close()



import traceback

from PySide2.QtCore import QObject, Signal, Slot
from twisted.internet.defer import CancelledError

from .errors import RefusedError, RespondError
from .file_transfer_protocol import FileTransferProtocol


class WormholeSignals(QObject):
    code_received = Signal(str)
    versions_received = Signal(dict)
    wormhole_open = Signal()
    wormhole_closed = Signal()
    wormhole_shutdown = Signal()
    wormhole_shutdown_received = Signal()
    message_sent = Signal(bool)
    message_received = Signal(str)
    file_receive_pending = Signal(str, int)
    file_transfer_progress = Signal(int, int, int)
    file_transfer_complete = Signal(int, str)
    error = Signal(Exception, str)
    respond_error = Signal(Exception, str)


class WormholeController:
    def __init__(self, reactor):
        super().__init__()
        self.signals = WormholeSignals()
        self._protocol = FileTransferProtocol(reactor, self.signals)

    @Slot(str)
    def open(self, code=None):
        self._capture_errors(self._protocol.open, code)

    @Slot(str)
    def set_code(self, code):
        @Slot(str)
        def open_with_code():
            self.signals.wormhole_closed.disconnect(open_with_code)
            self.open(code)

        self.signals.wormhole_closed.connect(open_with_code)
        self.close()

    @Slot()
    def close(self):
        self._capture_errors(self._protocol.close)

    @Slot()
    def shutdown(self):
        self._capture_errors(self._protocol.shutdown)

    @Slot()
    def send_message(self, message):
        self._capture_errors(self._protocol.send_message, message)

    @Slot(str, str)
    def send_file(self, id, file_path):
        self._capture_errors(self._protocol.send_file, id, file_path)

    @Slot(str, str)
    def receive_file(self, id, dest_path):
        self._capture_errors(self._protocol.receive_file, id, dest_path)

    @Slot()
    def reject_file(self):
        self.signals.respond_error.emit(
            RefusedError("The file was refused by the user"), None
        )

    def is_receiving_file(self):
        return self._protocol.is_receiving_file()

    def is_sending_file(self):
        return self._protocol.is_sending_file()

    def _capture_errors(self, command, *args, **kwds):
        try:
            command(*args, **kwds)
        except CancelledError:
            pass
        except RespondError as exception:
            self.signals.respond_error.emit(exception.cause, traceback.format.exc())
        except Exception as exception:
            self.signals.error.emit(exception, traceback.format_exc())



from PySide2.QtCore import Slot
from PySide2.QtWidgets import QDialog

from .ui_dialog import UiDialog


class ConnectDialog(UiDialog):
    def __init__(self, parent, wormhole):
        super().__init__(parent, "ConnectDialog.ui")

        self.wormhole = wormhole
        self.code = None

        self.set_code_button.clicked.connect(self._on_set_code_button)
        self.quit_button.clicked.connect(self.reject)

        wormhole.signals.wormhole_open.connect(self._on_wormhole_open)
        wormhole.signals.code_received.connect(self._on_code_received)
        wormhole.signals.error.connect(self._on_error)
        wormhole.signals.wormhole_closed.connect(self._on_wormhole_closed)

    def open(self):
        self._request_new_code()
        super().open()

    @Slot()
    def _on_wormhole_open(self):
        """
        Close the dialog if the wormhole is opened successfully.
        """
        self.accept()

    @Slot()
    def _on_wormhole_closed(self):
        """
        Open the dialog and attempt to repen the wormhole if the wormhole is closed.
        Do nothing if the dialog was manually closed.
        """
        if self.result() != QDialog.Rejected:
            self.open()

    @Slot(str)
    def _on_code_received(self, code):
        self.set_code_button.setEnabled(True)
        self.code = code[:100]
        self._refresh()

    @Slot()
    def _on_set_code_button(self):
        self.set_code_button.setDisabled(True)
        self.wormhole.set_code(self.code_edit.text().strip())

    @Slot(Exception, str)
    def _on_error(self, exception, traceback):
        self.set_code_button.setEnabled(True)

    def _request_new_code(self):
        # Don't allow code to be changed until one has been allocated
        self.code_edit.setText("")
        self.set_code_button.setDisabled(True)

        self.code = None
        self._refresh()
        self.wormhole.open()

    def _refresh(self):
        code = self.code
        if code is None:
            code = "[obtaining code...]"
        self.code_label.setText(code)



from collections import OrderedDict
from pathlib import Path

from PySide2.QtCore import Qt, Signal
from PySide2.QtWidgets import QHeaderView, QProgressBar, QTableWidget, QTableWidgetItem


class MessageTable(QTableWidget):
    send_file = Signal(int, str)

    def __init__(self, parent, wormhole):
        super().__init__(parent=parent)
        self.setAcceptDrops(True)

        self._send_files_pending = OrderedDict()
        self._wormhole = wormhole

        self._setup_columns()

    def _setup_columns(self):
        self.setColumnCount(2)
        header = self.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.Stretch)

    def add_sent_message(self, message):
        self._append_item(SendItem(f"You: {message}"))

    def add_received_message(self, message):
        self._append_item(ReceiveItem(message))

    def send_file_pending(self, filepath):
        id = self.rowCount()
        self._send_files_pending[id] = filepath
        self._append_item(SendFile(Path(filepath).name))

        if not self._wormhole.is_sending_file():
            self._send_next_file()

        return id

    def receiving_file(self, filepath):
        id = self.rowCount()
        self._append_item(ReceiveFile(Path(filepath).name))

        return id

    def transfer_progress(self, id, transferred_bytes, total_bytes):
        if not isinstance(self.cellWidget(id, 0), QProgressBar):
            bar = QProgressBar()
            bar.setTextVisible(False)
            bar.setMaximumWidth(32)
            self.setCellWidget(id, 0, bar)

        if total_bytes == 0:
            percent = 100
        else:
            percent = (100 * transferred_bytes) // total_bytes

        self.cellWidget(id, 0).setValue(percent)

    def transfer_complete(self, id, filename):
        self.item(id, 1).transfer_complete(filename)

        if not self._wormhole.is_sending_file():
            self._send_next_file()

    def _send_next_file(self):
        if self._send_files_pending:
            id, filepath = self._send_files_pending.popitem(last=False)
            self.item(id, 1).transfer_started()
            self.send_file.emit(id, filepath)

    def _append_item(self, item):
        item.setFlags(Qt.ItemIsEnabled)
        id = self.rowCount()
        self.insertRow(id)
        self.setItem(id, 1, item)
        self.resizeRowsToContents()

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            self.setStyleSheet("background-color: rgba(51, 153, 255, 0.2);")
            event.accept()

    def dragLeaveEvent(self, event):
        self.setStyleSheet("")
        event.accept()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()

    def dropEvent(self, event):
        self.setStyleSheet("")
        if event.mimeData().hasUrls:
            event.setDropAction(Qt.CopyAction)
            event.accept()

            for url in event.mimeData().urls():
                self.send_file_pending(url.toLocalFile())


class ReceiveItem(QTableWidgetItem):
    pass


class SendItem(QTableWidgetItem):
    def __init__(self, message):
        super().__init__(message)

        font = self.font()
        font.setItalic(True)
        self.setFont(font)


class ReceiveFile(ReceiveItem):
    def __init__(self, filename):
        self._filename = filename
        super().__init__(f"Receiving {self._filename}...")

    def transfer_complete(self, filename):
        self._filename = filename
        self.setText(f"Received {filename}")


class SendFile(SendItem):
    def __init__(self, filename):
        self._filename = filename
        super().__init__(f"Queued {self._filename}...")

    def transfer_started(self):
        self.setText(f"Sending {self._filename}...")

    def transfer_complete(self, filename):
        self._filename = filename
        self.setText(f"Sent {filename}")



import logging
import sys

from PySide2 import QtCore
from PySide2.QtWidgets import QApplication
import qt5reactor
import twisted.internet


# fix for pyinstaller packages app to avoid ReactorAlreadyInstalledError
# See https://github.com/kivy/kivy/issues/4182 and
# https://github.com/pyinstaller/pyinstaller/issues/3390
if "twisted.internet.reactor" in sys.modules:
    del sys.modules["twisted.internet.reactor"]

QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
QApplication([])
qt5reactor.install()

from .widgets.main_window import MainWindow  # noqa: E402
from .wormhole_controller import WormholeController  # noqa: E402


def run():
    logging.basicConfig(level=logging.INFO)

    reactor = twisted.internet.reactor
    wormhole = WormholeController(reactor)
    main_window = MainWindow(wormhole)
    main_window.run()

    sys.exit(reactor.run())



import logging

from PySide2.QtCore import Slot
from PySide2.QtWidgets import (
    QApplication,
    QDialog,
    QFileDialog,
    QMainWindow,
    QMessageBox,
)

from .connect_dialog import ConnectDialog
from .errors import get_error_text
from .message_table import MessageTable
from .save_file_dialog import SaveFileDialog
from .ui import CustomWidget, load_ui


class MainWindow(QMainWindow):
    def __init__(self, wormhole):
        super().__init__()
        load_ui(
            "MainWindow.ui",
            base_instance=self,
            custom_widgets=[CustomWidget(MessageTable, wormhole=wormhole)],
        )

        self.wormhole = wormhole

        self._hide_error()
        self.show()

    def run(self):
        self.connect_dialog = ConnectDialog(self, self.wormhole)
        self.save_file_dialog = SaveFileDialog(self)

        self.message_edit.returnPressed.connect(self.send_message_button.clicked)
        self.send_message_button.clicked.connect(self._on_send_message_button)
        self.send_files_button.clicked.connect(self._on_send_files_button)
        self.message_table.send_file.connect(self._on_send_file)

        self.connect_dialog.rejected.connect(self.close)

        self.save_file_dialog.finished.connect(self._on_save_file_dialog_finished)

        s = self.wormhole.signals
        s.wormhole_open.connect(self._hide_error)
        s.message_sent.connect(self._on_message_sent)
        s.message_received.connect(self._on_message_received)
        s.file_receive_pending.connect(self._on_file_receive_pending)
        s.file_transfer_progress.connect(self._on_file_transfer_progress)
        s.file_transfer_complete.connect(self._on_file_transfer_complete)
        s.error.connect(self._on_error)
        s.wormhole_shutdown_received.connect(self._on_wormhole_shutdown_received)
        s.wormhole_shutdown.connect(QApplication.quit)

        self.connect_dialog.open()

    def closeEvent(self, event):
        self.wormhole.signals.error.disconnect(self._on_error)
        self.wormhole.shutdown()

    @Slot()
    def _on_send_message_button(self):
        self._disable_message_entry()
        self.wormhole.send_message(self.message_edit.text())

    @Slot()
    def _on_send_files_button(self):
        dialog = QFileDialog(self, "Send")
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.fileSelected.connect(self._on_send_files_selected)
        dialog.open()

    @Slot(str)
    def _on_send_files_selected(self, filepath):
        if filepath != "":
            self.message_table.send_file_pending(filepath)

    @Slot(int, str)
    def _on_send_file(self, id, filepath):
        self.wormhole.send_file(id, filepath)

    @Slot()
    def _on_message_sent(self, success):
        self._enable_message_entry()
        if success:
            message = self.message_edit.text()
            self.message_table.add_sent_message(message)
            self.message_edit.clear()

    @Slot(str)
    def _on_message_received(self, message):
        self.message_table.add_received_message(message)

    @Slot(str, int)
    def _on_file_receive_pending(self, filename, size):
        self.save_file_dialog.open(filename, size)

    @Slot(int)
    def _on_save_file_dialog_finished(self, result):
        if result == QDialog.Accepted:
            id = self.message_table.receiving_file(self.save_file_dialog.filename)
            self.wormhole.receive_file(
                id, self.save_file_dialog.get_destination_directory()
            )
        else:
            self.wormhole.reject_file()

    @Slot(int, int, int)
    def _on_file_transfer_progress(self, id, transferred_bytes, total_bytes):
        self.message_table.transfer_progress(id, transferred_bytes, total_bytes)

    @Slot(int, str)
    def _on_file_transfer_complete(self, id, filename):
        self.message_table.transfer_complete(id, filename)

    @Slot(Exception, str)
    def _on_error(self, exception, traceback):
        logging.error(f"Caught Exception: {repr(exception)}")
        if traceback:
            logging.error(f"Traceback: {traceback}")

        self.error_label.setText(get_error_text(exception))
        self.error_label.show()

        self.wormhole.close()

    @Slot()
    def _hide_error(self):
        self.error_label.hide()

    @Slot()
    def _on_wormhole_shutdown_received(self):
        box = QMessageBox(parent=self)
        box.setText("The remote computer has closed the connection.")
        box.setIcon(QMessageBox.Information)
        box.setStandardButtons(QMessageBox.Close)
        box.setDefaultButton(QMessageBox.Close)
        box.exec_()

    def _disable_message_entry(self):
        self.message_edit.setDisabled(True)
        self.send_message_button.setDisabled(True)

    def _enable_message_entry(self):
        self.message_edit.setEnabled(True)
        self.send_message_button.setEnabled(True)



# Adapted from https://gist.github.com/cpbotha/1b42a20c8f3eb9bb7cb8
#
# Copyright (c) 2011 Sebastian Wiesner <lunaryorn@gmail.com>
# Modifications by Charl Botha <cpbotha@vxlabs.com>
# found this here:
# https://github.com/lunaryorn/snippets/blob/master/qt4/designer/pyside_dynamic.py
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from pathlib import Path

from PySide2.QtCore import QFile
from PySide2.QtUiTools import QUiLoader


class CustomWidget:
    def __init__(self, Widget, *args, **kwds):
        self.Widget = Widget
        self.name = self.Widget.__name__
        self.args = args
        self.kwds = kwds

    def create(self, parent):
        return self.Widget(parent, *self.args, **self.kwds)


class CustomUiLoader(QUiLoader):
    def __init__(self, base_instance=None, custom_widgets=None):
        super().__init__()
        self.base_instance = base_instance
        if custom_widgets is None:
            self.custom_widgets = {}
        else:
            self.custom_widgets = {w.name: w for w in custom_widgets}

    def createWidget(self, className, parent=None, name=""):
        if parent is None and self.base_instance:
            # No parent, this is the top-level widget
            return self.base_instance

        if className in QUiLoader.availableWidgets(self):
            widget = super().createWidget(className, parent, name)
        else:
            if className in self.custom_widgets:
                widget = self.custom_widgets[className].create(parent)
            else:
                raise KeyError("Unknown widget '%s'" % className)

        if self.base_instance:
            # Set an attribute for the new child widget on the base instance
            setattr(self.base_instance, name, widget)

        return widget


def load_ui(filename, base_instance=None, custom_widgets=None):
    ui_file = QFile(str(Path(__file__).parent / filename))
    ui_file.open(QFile.ReadOnly)

    loader = CustomUiLoader(base_instance, custom_widgets)
    ui = loader.load(ui_file)
    ui_file.close()

    return ui
